"""Base Class for all CalibrationInfo objects."""
import json
from pathlib import Path
from typing import Iterable, Tuple, Union, TypeVar

import numpy as np

CalInfo = TypeVar("CalInfo", bound="CalibrationInfo")


class CalibrationInfo:
    """Abstract BaseClass for all Calibration Info objects."""

    CAL_TYPE = None
    acc_unit = None
    gyro_unit = None

    _cal_type_explanation = """
    Note:
        All `CalibrationInfo` subclasses implement a `CAL_TYPE` attribute.
        If the calibration is exported into any format, this information is stored as well.
        If imported, all constructor methods intelligently infer the correct CalibrationInfo subclass based on
        this parameter.

        Example:
            >>> json_string = "{cal_type: 'Ferraris', ...}"
            >>> CalibrationInfo.from_json(json_string)
            <FerrarisCalibrationInfo ...>

    """

    __doc__ += _cal_type_explanation

    @property
    def _fields(self) -> Iterable[str]:
        """List of Calibration parameters that are required."""
        raise NotImplementedError("This method needs to be implemented by a subclass")

    def calibrate(self, acc: np.ndarray, gyro: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Abstract method to perform a calibration on both acc and gyro.

        This absolutely needs to implement by any daughter class

        Parameter
        ---------
        acc :
            3D acceleration
        gyro :
            3D gyroscope values

        """
        raise NotImplementedError("This method needs to be implemented by a subclass")

    def calibrate_gyro(self, gyro: np.ndarray) -> np.ndarray:
        """Abstract method to perform a calibration on both acc and gyro.

        This can implement by any daughter class, if separte calibration of acc makes sense for the calibration type.
        If not, an explicit error should be thrown

        Parameter
        ---------
        gro :
            3D gyroscope values

        """
        raise NotImplementedError("This method needs to be implemented by a subclass")

    def calibrate_acc(self, acc: np.ndarray) -> np.ndarray:
        """Abstract method to perform a calibration on the gyro.

        This can implement by any daughter class, if separte calibration of acc makes sense for the calibration type.
        If not, an explicit error should be thrown

        Parameter
        ---------
        acc :
            3D acceleration

        """
        raise NotImplementedError("This method needs to be implemented by a subclass")

    def __init__(self, **kwargs):
        """Create new CalibrationInfo instance.

        Parameter
        ---------
        kwargs :
            Matrices for all fields specified in `self._fields`

        """
        for field in self._fields:
            setattr(self, field, kwargs.get(field, None))

    def __repr__(self):
        """Show a proper string representation of a CalibrationInfo object."""
        out = self.__class__.__name__ + "("
        for val in self._fields:
            out += "\n" + val + " =\n" + getattr(self, val).__repr__() + ",\n"
        out += "\n)"
        return out

    def __eq__(self, other):
        """Check if two calibrations are identical."""
        # Check type:
        if not isinstance(other, self.__class__):
            raise ValueError("Comparison is only defined between two {} object!".format(self.__class__.__name__))

        # Test keys equal:
        if not self._fields == other._fields:
            return False

        # Test method equal
        if not self.CAL_TYPE == other.CAL_TYPE:
            return False

        # Test units
        if (not self.acc_unit == other.acc_unit) or (not self.gyro_unit == other.gyro_unit):
            return False

        # Test Calibration values
        for v1, v2 in zip(self.__dict__.values(), other.__dict__.values()):
            if not np.array_equal(v1, v2):
                return False
        return True

    def _to_list_dict(self):
        d = {key: getattr(self, key).tolist() for key in self._fields}
        d["cal_type"] = self.CAL_TYPE
        return d

    @classmethod
    def _from_list_dict(cls, list_dict):
        raw_json = {k: np.array(v) for k, v in list_dict.items()}
        return cls(**raw_json)

    @classmethod
    def _get_subclasses(cls):
        for subclass in cls.__subclasses__():
            yield from subclass._get_subclasses()
            yield subclass

    @classmethod
    def find_subclass_from_cal_type(cls, cal_type):
        """Get a SensorCalibration subclass that handles the specified calibration type."""
        return next(x for x in CalibrationInfo._get_subclasses() if x.CAL_TYPE == cal_type)

    def to_json(self) -> str:
        """Convert all calibration matrices into a json string."""
        data_dict = self._to_list_dict()
        return json.dumps(data_dict, indent=4)

    @classmethod
    def from_json(cls, json_str: str) -> CalInfo:
        """Create a calibration object from a json string (created by `CalibrationInfo.to_json`).

        Parameter
        ---------
        json_str: valid json string object

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        raw_json = json.loads(json_str)
        subclass = cls.find_subclass_from_cal_type(raw_json["cal_type"])
        return subclass._from_list_dict(raw_json)

    def to_json_file(self, path: Union[str, Path]):
        """Dump acc calibration matrices into a file in json format.

        Parameter
        ---------
        path :
            path to the json file

        """
        data_dict = self._to_list_dict()
        return json.dump(data_dict, open(path, "w"))

    @classmethod
    def from_json_file(cls, path: Union[str, Path]) -> CalInfo:
        """Create a calibration object from a valid json file (created by `CalibrationInfo.to_json_file`).

        Parameter
        ---------
        path :
            Path to the json file

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        raw_json = json.load(open(path, "r"))
        subclass = cls.find_subclass_from_cal_type(raw_json["cal_type"])
        return subclass._from_list_dict(raw_json)

    def to_hdf5(self, path: Union[str, Path]):
        """Save calibration matrices to hdf5 file format.

        Parameter
        ---------
        path :
            Path to the hdf5 file

        """
        import h5py  # noqa: import-outside-toplevel

        with h5py.File(path, "w") as hdf:
            d = {key: getattr(self, key).tolist() for key in self._fields}
            for k, v in d.items():
                hdf.create_dataset(k, data=v)
            hdf["cal_type"] = self.CAL_TYPE
            hdf["acc_unit"] = self.acc_unit
            hdf["gyro_unit"] = self.gyro_unit

    @classmethod
    def from_hdf5(cls, path: Union[str, Path]):
        """Read calibration data stored in hdf5 fileformat (created by `CalibrationInfo.save_to_hdf5`).

        Parameter
        ---------
        path :
            Path to the hdf5 file

        Returns
        -------
        cal_info
            A CalibrationInfo object.
            The exact child class is determined by the `cal_type` key in the json string.

        """
        import h5py  # noqa: import-outside-toplevel

        with h5py.File(path, "r") as hdf:
            values = dict()
            subcls = cls.find_subclass_from_cal_type(hdf["cal_type"][...])
            for k in subcls._fields:
                values[k] = np.array(hdf.get(k))

        return subcls(**values)



"""Calculate a Ferraris calibration from sensor data."""
from itertools import product
from typing import Optional, Iterable, TypeVar, Type, Tuple

import numpy as np
import pandas as pd
from numpy.linalg import inv

from imucal.calibration_gui import _convert_data_from_section_list_to_df, CalibrationGui
from imucal.calibration_info import CalibrationInfo
from imucal.ferraris_calibration_info import FerrarisCalibrationInfo, TurntableCalibrationInfo

T = TypeVar("T", bound="FerrarisCalibration")


class FerrarisCalibration:
    """Calculate a Ferraris calibration matrices based on a set of calibration movements.

    The Ferraris calibration is derived based on a well defined series of data recordings:

    `x_p`: positive x-axis of sensor is aligned with gravity (x-acc measures +1g)
    `x_a`: negative x-axis of sensor is aligned with gravity (x-acc measures -1g)
    `y_p`: positive y-axis of sensor is aligned with gravity (y-acc measures +1g)
    `y_a`: negative y-axis of sensor is aligned with gravity (y-acc measures -1g)
    `z_p`: positive z-axis of sensor is aligned with gravity (z-acc measures +1g)
    `z_a`: negative z-axis of sensor is aligned with gravity (z-acc measures -1g)

    `x_rot`: sensor is rotated clockwise in the `x_p` position around the x-axis (x-gyro shows negative values) for a
        well known angle (typically 360 deg)
    `y_rot`: sensor is rotated clockwise in the `y_p` position around the y-axis (y-gyro shows negative values) for a
        well known angle (typically 360 deg)
    `z_rot`: sensor is rotated clockwise in the `z_p` position around the z-axis (z-gyro shows negative values) for a
        well known angle (typically 360 deg)

    All sections need to be recorded for a sensor and then annotated.
    In particular for the rotation, it is important to annotate the data directly at the end and the beginning of the
    rotation to avoid noise and artifact degrading the integration results.
    This class offers various helper constructors to support the annotation process.


    Notes
    -----
    Depending on how the axis of your respective sensor coordinate system are defined and how you perform the
    calibration, you might need to change the `grav` and `expected_angle` parameter.

    Typical situations are:
        - If you define the positive axis direction as the direction, where the acc measures -g, change `grav` to
            -9.81 m/s^2
        - If you perform a counter-clockwise rotation during the calibration, set `expected_angle` to +360
        - For combinations of both, both parameter might need to be adapted


    Examples
    --------
    >>> from imucal import FerrarisCalibration
    >>> sampling_rate = 100 #Hz
    >>> data = ... # my data as 6 col pandas dataframe
    >>> # This will open an interactive plot, where you can select the start and the stop sample of each region
    >>> cal, section_list = FerrarisCalibration.from_interactive_plot(data, sampling_rate=sampling_rate)
    >>> section_list.to_csv('./calibration_sections.csv')  # This is optional, but recommended
    >>> calibration_info = cal.compute_calibration_matrix()  # Calculate the actual matrizes.
    >>> calibration_info
    < FerrarisCalibration object at ... >

    """

    acc_x_p: np.ndarray
    acc_x_a: np.ndarray
    acc_y_p: np.ndarray
    acc_y_a: np.ndarray
    acc_z_p: np.ndarray
    acc_z_a: np.ndarray
    gyr_x_p: np.ndarray
    gyr_x_a: np.ndarray
    gyr_y_p: np.ndarray
    gyr_y_a: np.ndarray
    gyr_z_p: np.ndarray
    gyr_z_a: np.ndarray

    acc_x_rot: np.ndarray
    acc_y_rot: np.ndarray
    acc_z_rot: np.ndarray
    gyr_x_rot: np.ndarray
    gyr_y_rot: np.ndarray
    gyr_z_rot: np.ndarray

    sampling_rate: float
    grav: float
    expected_angle: float

    EXPECTED_ANGLE: float = -360.0  # +360 for anti-clockwise rotation, -360 for clockwise rotation
    DEFAULT_GRAV: float = 9.81
    ACC_COLS = ("acc_x", "acc_y", "acc_z")
    GYRO_COLS = ("gyr_x", "gyr_y", "gyr_z")
    FERRARIS_SECTIONS = ("x_p", "x_a", "y_p", "y_a", "z_p", "z_a", "x_rot", "y_rot", "z_rot")

    _CALIBRATION_INFO = FerrarisCalibrationInfo

    _fields = tuple("{}_{}".format(x, y) for x, y in product(("acc", "gyr"), FERRARIS_SECTIONS))

    def __init__(
        self, sampling_rate: float, grav: Optional[float] = None, expected_angle: Optional[float] = None, **kwargs
    ) -> None:
        """Create a Calibration object.

        Usually you would want to use one of the provided "from" constructors instead of directly using the init.

        Parameter
        ---------
        sampling_rate :
            Sampling rate of the data
        expected_angle :
            expected rotation angle for the gyroscope rotation.
            If None defaults to `FerrarisCalibration.EXPECTED_ANGLE`
        grav :
            The expected value of the gravitational acceleration.
            Defaults to `FerrarisCalibration.DEFAULT_GRAV`
        kwargs :
            A 3D numpy array for each section (acc and gyro separately) required by the ferraris calibration.
            The arguments need to be named acc/gyr_section (e.g. acc_x_p)

        """
        for field in self._fields:
            setattr(self, field, kwargs.get(field, None))

        self.sampling_rate = sampling_rate
        self.grav = grav or self.DEFAULT_GRAV
        self.expected_angle = expected_angle or self.EXPECTED_ANGLE
        super().__init__()

    @classmethod
    def from_df(
        cls: Type[T],
        df: pd.DataFrame,
        sampling_rate: float,
        grav: Optional[float] = None,
        expected_angle: Optional[float] = None,
        acc_cols: Optional[Iterable[str]] = None,
        gyro_cols: Optional[Iterable[str]] = None,
    ) -> T:
        """Create a Calibration object based on a dataframe which has all required sections labeled.

        The expected Dataframe has the section label as index and has at least the 6 required data columns.
        The index must contain all sections as specified by `FerrarisCalibration.FERRARIS_SECTIONS`.

        >>> print(example_df)
                acc_x acc_y   acc_z  gyr_x  gyr_y  gyr_z
        part
        x_a   -2052.0 -28.0   -73.0    1.0    0.0   -5.0
        x_a   -2059.0 -29.0   -77.0    2.0   -3.0   -5.0
        x_a   -2054.0 -25.0   -71.0    3.0   -2.0   -3.0
        ...       ...   ...     ...    ...    ...    ...
        z_rot   -36.0  35.0  2079.0    2.0   -5.0   -2.0
        z_rot   -28.0  36.0  2092.0    6.0   -5.0   -4.0
        z_rot   -36.0  21.0  2085.0    5.0   -4.0   -4.0

        Examples
        --------
        >>> from imucal import FerrarisCalibration
        >>> import pandas as pd
        >>> sampling_rate = 100 #Hz
        >>> df = ... # A valid DataFrame with all sections in the index
        >>> cal = FerrarisCalibration.from_dict(df, sampling_rate=sampling_rate)
        < FerrarisCalibration object at ... >

        Parameter
        ---------
        df :
            6 column dataframe (3 acc, 3 gyro)
        sampling_rate :
            Sampling rate of the data
        expected_angle :
            expected rotation angle for the gyroscope rotation.
            If None defaults to `FerrarisCalibration.EXPECTED_ANGLE`
        grav :
            The expected value of the gravitational acceleration.
            Defaults to `FerrarisCalibration.DEFAULT_GRAV`
        acc_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.ACC_COLS`
        gyro_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.GYRO_COLS`

        Returns
        -------
        ferraris_cal_obj : FerrarisCalibration

        """
        if acc_cols is None:
            acc_cols = list(cls.ACC_COLS)
        if gyro_cols is None:
            gyro_cols = list(cls.GYRO_COLS)

        acc_df = df[list(acc_cols)]
        gyro_df = df[list(gyro_cols)]
        acc_dict = acc_df.groupby(level=0).apply(lambda x: x.values).to_dict()
        gyro_dict = gyro_df.groupby(level=0).apply(lambda x: x.values).to_dict()
        acc_dict = {"acc_" + k: v for k, v in acc_dict.items()}
        gyro_dict = {"gyr_" + k: v for k, v in gyro_dict.items()}

        return cls(sampling_rate, grav, expected_angle, **acc_dict, **gyro_dict)

    @classmethod
    def from_section_list(
        cls: Type[T],
        data: pd.DataFrame,
        section_list: pd.DataFrame,
        sampling_rate: float,
        grav: Optional[float] = None,
        expected_angle: Optional[float] = None,
        acc_cols: Optional[Iterable[str]] = None,
        gyro_cols: Optional[Iterable[str]] = None,
    ) -> T:
        """Create a Calibration object based on a valid section list.

        A section list marks the start and the endpoints of each required section in the data object.
        A valid section list is usually created using `FerrarisCalibration.from_interactive_plot()`.
        This section list can be stored on disk and this method can be used to turn it back into a valid calibration
        object.

        Parameter
        ---------
        df :
            6 column dataframe (3 acc, 3 gyro)
        section_list :
            A pandas dataframe representing a section list
        sampling_rate :
            Sampling rate of the data
        expected_angle :
            expected rotation angle for the gyroscope rotation.
            If None defaults to `FerrarisCalibration.EXPECTED_ANGLE`
        grav :
            The expected value of the gravitational acceleration.
            Defaults to `FerrarisCalibration.DEFAULT_GRAV`
        acc_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.ACC_COLS`
        gyro_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.GYRO_COLS`

        Returns
        -------
        ferraris_cal_obj : FerrarisCalibration


        Examples
        --------
        >>> from imucal import FerrarisCalibration
        >>> import pandas as pd
        >>> # Load a valid section list from disk. Note the `index_col=0` to preserve correct format!
        >>> section_list = pd.read_csv('./calibration_sections.csv', index_col=0)
        >>> sampling_rate = 100 #Hz
        >>> data = ... # my data as 6 col pandas dataframe
        >>> cal = FerrarisCalibration.from_section_list(data, section_list, sampling_rate=sampling_rate)
        < FerrarisCalibration object at ... >

        """
        df = _convert_data_from_section_list_to_df(data, section_list)
        return cls.from_df(
            df, sampling_rate, expected_angle=expected_angle, grav=grav, acc_cols=acc_cols, gyro_cols=gyro_cols
        )

    @classmethod
    def from_interactive_plot(
        cls: Type[T],
        data: pd.DataFrame,
        sampling_rate: float,
        expected_angle: Optional[float] = None,
        grav: Optional[float] = None,
        acc_cols: Optional[Iterable[str]] = None,
        gyro_cols: Optional[Iterable[str]] = None,
    ) -> Tuple[T, pd.DataFrame]:
        """Create a Calibration object by selecting the individual signal sections manually in an interactive GUI.

        This will open a Tkinter Window that allows you to label the start and the end all required sections for a
        Ferraris Calibration.
        See the class docstring for more detailed explanations of these sections.

        Examples
        --------
        >>> from imucal import FerrarisCalibration
        >>> sampling_rate = 100 #Hz
        >>> data = ... # my data as 6 col pandas dataframe
        >>> # This will open an interactive plot, where you can select the start and the stop sample of each region
        >>> cal, section_list = FerrarisCalibration.from_interactive_plot(data, sampling_rate=sampling_rate)
        >>> section_list.to_csv('./calibration_sections.csv')  # This is optional, but recommended
        >>> cal
        < FerrarisCalibration object at ... >

        Parameter
        ---------
        df :
            6 column dataframe (3 acc, 3 gyro)
        sampling_rate :
            Sampling rate of the data
        expected_angle :
            expected rotation angle for the gyroscope rotation.
            If None defaults to `FerrarisCalibration.EXPECTED_ANGLE`
        grav :
            The expected value of the gravitational acceleration.
            Defaults to `FerrarisCalibration.DEFAULT_GRAV`
        acc_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.ACC_COLS`
        gyro_cols :
            The name of the 3 acceleration columns in order x,y,z.
            Defaults to `FerrarisCalibration.GYRO_COLS`

        Returns
        -------
        ferraris_cal_obj : FerrarisCalibration
        section_list : pd.DataFrame
            Section list representing the start and stop of each section.
            It is advised to save this to disk to avoid repeated manual labeling.
            `FerrarisCalibration.from_section_list()` can be used to recreate the calibration object

        """
        if acc_cols is None:
            acc_cols = list(cls.ACC_COLS)
        if gyro_cols is None:
            gyro_cols = list(cls.GYRO_COLS)

        acc = data[acc_cols].values
        gyro = data[gyro_cols].values

        section_list = _find_calibration_sections_interactive(acc, gyro)
        return (
            cls.from_section_list(
                data,
                section_list,
                sampling_rate,
                expected_angle=expected_angle,
                grav=grav,
                acc_cols=acc_cols,
                gyro_cols=gyro_cols,
            ),
            section_list,
        )

    def compute_calibration_matrix(self) -> CalibrationInfo:
        """Compute the calibration Information.

        This actually performs the Ferraris calibration following the original publication equation by equation.
        """
        cal_mat = self._CALIBRATION_INFO()

        ###############################################################################################################
        # Compute Acceleration Matrix

        # Calculate means from all static phases and stack them into 3x3 matrices
        # Note: Each measurement should be a column
        U_a_p = np.vstack(  # noqa: invalid-name
            (np.mean(self.acc_x_p, axis=0), np.mean(self.acc_y_p, axis=0), np.mean(self.acc_z_p, axis=0),)
        ).T
        U_a_n = np.vstack(  # noqa: invalid-name
            (np.mean(self.acc_x_a, axis=0), np.mean(self.acc_y_a, axis=0), np.mean(self.acc_z_a, axis=0),)
        ).T

        # Eq. 19
        U_a_s = U_a_p + U_a_n  # noqa: invalid_name

        # Bias Matrix
        # Eq. 20
        B_a = U_a_s / 2  # noqa: invalid_name

        # Bias Vector
        b_a = np.diag(B_a)
        cal_mat.b_a = b_a

        # Compute Scaling and Rotation
        # No need for bias correction, since it cancels out!
        # Eq. 21
        U_a_d = U_a_p - U_a_n  # noqa: invalid_name

        # Calculate Scaling matrix
        # Eq. 23
        k_a_sq = 1 / (4 * self.grav ** 2) * np.diag(U_a_d @ U_a_d.T)
        K_a = np.diag(np.sqrt(k_a_sq))  # noqa: invalid_name
        cal_mat.K_a = K_a  # noqa: invalid_name

        # Calculate Rotation matrix
        # Eq. 22
        R_a = inv(K_a) @ U_a_d / (2 * self.grav)  # noqa: invalid_name
        cal_mat.R_a = R_a  # noqa: invalid_name

        ###############################################################################################################
        # Calculate Gyroscope Matrix

        # Gyro Bias from the static phases of the acc calibration
        # One static phase would be sufficient, but why not use all of them if you have them.
        # Note that this calibration ignores any influences due to the earth rotation.

        b_g = np.mean(
            np.vstack((self.gyr_x_p, self.gyr_x_a, self.gyr_y_p, self.gyr_y_a, self.gyr_z_p, self.gyr_z_a,)), axis=0,
        )

        cal_mat.b_g = b_g

        # Acceleration sensitivity

        # Note: Each measurement should be a column
        U_g_p = np.vstack(  # noqa: invalid_name
            (np.mean(self.gyr_x_p, axis=0), np.mean(self.gyr_y_p, axis=0), np.mean(self.gyr_z_p, axis=0),)
        ).T
        U_g_a = np.vstack(  # noqa: invalid_name
            (np.mean(self.gyr_x_a, axis=0), np.mean(self.gyr_y_a, axis=0), np.mean(self.gyr_z_a, axis=0),)
        ).T

        # Eq. 9
        K_ga = (U_g_p - U_g_a) / (2 * self.grav)  # noqa: invalid_name
        cal_mat.K_ga = K_ga  # noqa: invalid_name

        # Gyroscope Scaling and Rotation

        # First apply partial calibration to remove offset and acc influence
        acc_x_rot_cor = cal_mat.calibrate_acc(self.acc_x_rot)
        acc_y_rot_cor = cal_mat.calibrate_acc(self.acc_y_rot)
        acc_z_rot_cor = cal_mat.calibrate_acc(self.acc_z_rot)
        gyr_x_rot_cor = cal_mat._calibrate_gyro_offsets(self.gyr_x_rot, acc_x_rot_cor)
        gyr_y_rot_cor = cal_mat._calibrate_gyro_offsets(self.gyr_y_rot, acc_y_rot_cor)
        gyr_z_rot_cor = cal_mat._calibrate_gyro_offsets(self.gyr_z_rot, acc_z_rot_cor)

        # Integrate gyro readings
        # Eg. 13/14
        W_s = np.zeros((3, 3))  # noqa: invalid_name
        W_s[:, 0] = np.sum(gyr_x_rot_cor, axis=0) / self.sampling_rate
        W_s[:, 1] = np.sum(gyr_y_rot_cor, axis=0) / self.sampling_rate
        W_s[:, 2] = np.sum(gyr_z_rot_cor, axis=0) / self.sampling_rate

        # Eq.15
        expected_angles = self.expected_angle * np.identity(3)
        multiplied = W_s @ inv(expected_angles)

        # Eq. 12
        k_g_sq = np.diag(multiplied @ multiplied.T)
        K_g = np.diag(np.sqrt(k_g_sq))  # noqa: invalid_name
        cal_mat.K_g = K_g  # noqa: invalid_name

        R_g = inv(K_g) @ multiplied  # noqa: invalid_name
        cal_mat.R_g = R_g  # noqa: invalid_name

        return cal_mat


class TurntableCalibration(FerrarisCalibration):
    """Calculate a Ferraris calibration matrices based on a turntable measurement.

    This calibration is basically identical to the FerrarisCalibration.
    However, the calibrate method will return a `TurntableCalibrationInfo` to indicate the expected higher precision
    of this calibration method.

    Further this Calibration expects rotations of 720 deg by default, as this is common for many turntables.
    For further information on the sign of the expected rotation angle see the `FerrarisCalibration`.
    """

    _CALIBRATION_INFO = TurntableCalibrationInfo

    EXPECTED_ANGLE: float = -720.0


def _find_calibration_sections_interactive(acc: np.ndarray, gyro: np.ndarray, title: Optional[str] = None):
    """Prepare the calibration data for the later calculation of calibration matrices.

    Parameters
    ----------
    acc : (n, 3) array
        Acceleration data
    gyro : (n, 3) array
        Gyroscope data
    title :
        Optional title for the Calibration GUI

    """
    plot = CalibrationGui(acc, gyro, FerrarisCalibration.FERRARIS_SECTIONS, title=title)

    section_list = plot.section_list

    check_all = (all(v) for v in section_list.values())
    if not all(check_all):
        raise ValueError("Some regions are missing in the section list. Label all regions before closing the plot")

    section_list = pd.DataFrame(section_list, index=("start", "end")).T

    return section_list



from imucal.calibration_info import CalibrationInfo  # noqa: F401
from imucal.ferraris_calibration import FerrarisCalibration, TurntableCalibration  # noqa: F401
from imucal.ferraris_calibration_info import FerrarisCalibrationInfo, TurntableCalibrationInfo  # noqa: F401

__all__ = [
    "CalibrationInfo",
    "FerrarisCalibration",
    "TurntableCalibration",
    "FerrarisCalibrationInfo",
    "TurntableCalibrationInfo",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imucal']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.2,<2.0.0', 'pandas>=1.1.3,<2.0.0']

extras_require = \
{'calplot': ['matplotlib>=3.3.2,<4.0.0'], 'h5py': ['h5py>=2.10.0,<3.0.0']}

setup_kwargs = {
    'name': 'imucal',
    'version': '0.1.0',
    'description': 'A Python library to calibrate 6 DOF IMUs',
    'long_description': "# IMU Calibration\n\nThis package provides methods to calculate and apply calibrations for IMUs based on multiple different methods.\n\nSo far supported are:\n\n- Ferraris Calibration (Ferraris1995)\n- Ferraris Calibration using a Turntable\n\n## Installation\n\n```\npip install imucal\n```\n\nTo use the included calibration GUI you also need matplotlib (version >2.2).\nYou can install it using:\n\n```\npip install imucal[calplot]\n```\n\n## Calibrations\n\n### Ferraris\n\nThis package implements the IMU-infield calibration based on Ferraris1995.\nThis calibration methods requires the IMU data from 6 static positions (3 axis parallel and antiparallel to gravitation vector) and 3 rotations around the 3 main axis.\nIn this implementation these parts are referred to as follows `{acc,gry}_{x,y,z}_{p,a}` for the static regions and `{acc,gry}_{x,y,z}_rot` for the rotations.\nAs example, `acc_y_a` would be the 3D-acceleration data measured during a static phase, where the **y-axis** was oriented **antiparallel** to the gravitation.\n\n#### Creating a new Calibration Object\n\nIf the data of all of these sections is already available separately as numpy arrays of the shape `(n x 3)`, where `n` is the number of samples in each section, they can be directly used to initialize a Calibration object:\n\n```python\nfrom imucal import FerrarisCalibration\n\nsampling_rate = 100 #Hz \ncal = FerrarisCalibration(sampling_rate=sampling_rate,\n                          acc_x_p=data_acc_x_p,\n                          ...,\n                          gyr_z_rot=data_gyr_x_rot)\n```\nor via class variables:\n```python\nfrom imucal import FerrarisCalibration\n\nsampling_rate = 100 #Hz \ncal = FerrarisCalibration(sampling_rate=sampling_rate)\ncal.gyr_z_a = data_gyr_z_a\n...\n```\n\nIf the data was recorded as a single continuous stream, we first need to identify the different regions in the data.\nThe `from_interactive_plot` method can be used to extract them manually using a GUI.\nFor this and for most other methods, we expect the data to be a `pd.DataFrame` with the columns `'acc_x', 'acc_y', 'acc_z, 'gyr_x', 'gyr_y', 'gyr_z'`, where each column represents the datastream of one sensor axis.\n\nNote: The expected column names can be overwritten\n\n```python\nfrom imucal import FerrarisCalibration\n\nsampling_rate = 100 #Hz \n# This will open an interactive plot, where you can select the start and the stop sample of each region\ncal, section_list = FerrarisCalibration.from_interactive_plot(data, sampling_rate=sampling_rate)\n\nsection_list.to_csv('./calibration_sections.csv')  # This is optional, but recommended\n```\n\nThis method also returns the `section_list` in addition to the Calibration object.\nIt is advised to save this list, as it can be used recreate the Calibration without performing the manual selection of the regions again:\n\n```python\n# At some other day:\nfrom imucal import FerrarisCalibration\nimport pandas as pd\n\nsection_list = pd.read_csv('./calibration_sections.csv', index_col=0)\nsampling_rate = 100 #Hz \n# This will recreate the calibration\ncal = FerrarisCalibration.from_section_list(data, section_list, sampling_rate=sampling_rate)\n```\n\n#### Performing the Calibration\n\nWhen the calibration object was successful initialized, you can obtain the calibration by simply calling `compute_calibration_matrix`:\n\n```python\ncal_mat = cal.compute_calibration_matrix()\nprint(cal_mat)\n```\n\nThis will return an `FerrarisCalibrationInfo` object, which holds all required calibration information.\nThis object can be saved and loaded to and from `hdf5` and `json` using the respective methods:\n\n```python\ncal_mat.to_json_file('./calibration.json')\n# some other day:\nfrom imucal import FerrarisCalibrationInfo\n\ncal_mat = FerrarisCalibrationInfo.from_json_file('./calibration.json') \n```\n\n```python\ncal_mat.to_hdf5('./calibration.h5')\n# some other day:\nfrom imucal import FerrarisCalibrationInfo\n\ncal_mat = FerrarisCalibrationInfo.from_hdf5('./calibration.h5') \n```\n\n#### Applying the Calibration\n\nThe `FerrarisCalibrationInfo` object can be used to apply the Calibration to new data from the same sensor:\n\n```python\ncalibrated_acc, calibrated_gyro = cal_mat.calibrate(acc, gyro)\n```\n\n`acc` and `gyro` are expected to be numpy arrays in the shape (n x 3)\n\n\n### Turntable Calibration\n\nThe turntable calibration uses some sort of instumeted turntable to perform the orientation changes and rotations for the Ferraris calibration.\nIt is fundamentally identical to the simple Ferraris calibration, however to indicate the expected difference in precision, this calibration method is implemented as seperate classes.\nThe interface is identical to the Ferraris calibration and all methods shown above can be used.\n\n**Note**: By default the Turntable calibration expects 720 deg rotations instead of 360 deg\n\n```python\n# obtain the calibration\nfrom imucal import TurntableCalibration\n\nsampling_rate = 100 #Hz \n# This will open an interactive plot, where you can select the start and the stop sample of each region\ncal, section_list = TurntableCalibration.from_interactive_plot(data, sampling_rate=sampling_rate)\ncal_mat = cal.compute_calibration_matrix()\ncal_mat.to_json_file('./calibration.json')\n\n# Apply the calibration\nfrom imucal import TurntableCalibrationInfo\n\ncal_mat = TurntableCalibrationInfo.from_json_file('./calibration.json') \n\ncalibrated_acc, calibrated_gyro = cal_mat.calibrate(acc, gyro)\n```",
    'author': 'Arne Küderle',
    'author_email': 'arne.kuederle@fau.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mad-lab-fau/imucal',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Wrapper object to hold calibration matrices for a Ferraris Calibration."""
import warnings
from typing import Tuple

import numpy as np

from imucal.calibration_info import CalibrationInfo


# TODO: Add example to docstring
class FerrarisCalibrationInfo(CalibrationInfo):
    """Calibration object that represents all the required information to apply a Ferraris calibration to a dataset.

    Parameters
    ----------
    K_a :
        Scaling matrix for the acceleration
    R_a :
        Rotation matrix for the acceleration
    b_a :
        Acceleration bias
    K_g :
        Scaling matrix for the gyroscope
    R_g :
        Rotation matrix for the gyroscope
    K_ga :
        Influence of acceleration on gyroscope
    b_g :
        Gyroscope bias

    """

    CAL_TYPE = "Ferraris"
    acc_unit = "m/s^2"
    gyro_unit = "deg/s"
    K_a: np.ndarray
    R_a: np.ndarray
    b_a: np.ndarray
    K_g: np.ndarray
    R_g: np.ndarray
    K_ga: np.ndarray
    b_g: np.ndarray

    _fields = ("K_a", "R_a", "b_a", "K_g", "R_g", "K_ga", "b_g")

    __doc__ += CalibrationInfo._cal_type_explanation

    def calibrate_acc(self, acc: np.ndarray) -> np.ndarray:
        """Calibrate the accelerometer.

        This corrects scaling, rotation, non-orthogonalities and bias.

        Parameters
        ----------
        acc :
            3D acceleration

        Returns
        -------
        Calibrated acceleration

        """
        # Check if all required paras are initialized to throw appropriate error messages:
        paras = ("K_a", "R_a", "b_a")
        for v in paras:
            if getattr(self, v, None) is None:
                raise ValueError(
                    "{} need to initialised before an acc calibration can be performed. {} is missing".format(paras, v)
                )

        # Combine Scaling and rotation matrix to one matrix
        acc_mat = np.linalg.inv(self.R_a) @ np.linalg.inv(self.K_a)
        acc_out = acc_mat @ (acc - self.b_a).T

        return acc_out.T

    def calibrate_gyro(self, gyro: np.ndarray) -> np.ndarray:
        """Calibrate the gyroscope.

        .. warning ::
            This is not supported for the FerrarisCalibration, as it is not possible to fully calibrate the gyroscope
            without the acc values.
            Any acc interference on the gyroscope (`K_ga`) will not be taken into account.
            Use `FerrarisCalibrationInfo.calibrate` instead.

        Parameters
        ----------
        gyro :
            3D gyroscope values

        Warns
        -----
        UserWarning
            Always, informing about the missing `K_ga` calibration

        """
        warnings.warn(
            "Performing a calibration on the Gyro data alone, will not correct potential acc-gyro"
            " interferences. Use `{}CalibrationInfo.calibrate` to calibrate acc and gyro"
            " together.".format(self.CAL_TYPE)
        )
        return self._calibrate_gyro(gyro, calibrated_acc=None)

    def calibrate(self, acc: np.ndarray, gyro: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Calibrate the accelerometer and the gyroscope.

        This corrects:
            acc: scaling, rotation, non-orthogonalities, and bias
            gyro: scaling, rotation, non-orthogonalities, bias, and acc influence on gyro

        Parameters
        ----------
        acc :
            3D acceleration
        gyro :
            3D gyroscope values

        Returns
        -------
        Corrected acceleration and gyroscope values

        """
        # Check if all required paras are initialized to throw appropriate error messages:
        for v in self._fields:
            if getattr(self, v, None) is None:
                raise ValueError(
                    "{} need to initialised before an acc calibration can be performed. {} is missing".format(
                        self._fields, v
                    )
                )

        acc_out = self.calibrate_acc(acc)
        gyro_out = self._calibrate_gyro(gyro, acc_out)

        return acc_out, gyro_out

    def _calibrate_gyro(self, gyro, calibrated_acc=None):
        # Check if all required paras are initialized to throw appropriate error messages:
        required = ["K_g", "R_g", "b_g"]
        if calibrated_acc is not None:
            required += ["K_ga"]
        for v in required:
            if getattr(self, v, None) is None:
                raise ValueError(
                    "{} need to initialised before an gyro calibration can be performed. {} is missing".format(
                        required, v
                    )
                )
        # Combine Scaling and rotation matrix to one matrix
        gyro_mat = np.matmul(np.linalg.inv(self.R_g), np.linalg.inv(self.K_g))
        tmp = self._calibrate_gyro_offsets(gyro, calibrated_acc)

        gyro_out = gyro_mat @ tmp.T
        return gyro_out.T

    def _calibrate_gyro_offsets(self, gyro, calibrated_acc=None):
        if calibrated_acc is None:
            d_ga = np.array(0)
        else:
            d_ga = self.K_ga @ calibrated_acc.T
        offsets = d_ga.T + self.b_g
        return gyro - offsets


class TurntableCalibrationInfo(FerrarisCalibrationInfo):
    """Calibration object that represents all the required information to apply a Turntable calibration to a dataset.

    A Turntable calibration is identical to a Ferraris calibration.
    However, because the parameters are calculated using a calibration table instead of hand rotations,
    higher precision is expected.

    Parameters
    ----------
    K_a :
        Scaling matrix for the acceleration
    R_a :
        Rotation matrix for the acceleration
    b_a :
        Acceleration bias
    K_g :
        Scaling matrix for the gyroscope
    R_g :
        Rotation matrix for the gyroscope
    K_ga :
        Influence of acceleration on gyroscope
    b_g :
        Gyroscope bias

    """

    CAL_TYPE = "Turntable"
